# This file is part of dh-cmake, and is distributed under the OSI-approved
# BSD 3-Clause license. See top-level LICENSE file or
# https://gitlab.kitware.com/debian/dh-cmake/blob/master/LICENSE for details.

import re
from dhcmake import common


class DHCMake(common.DHCommon):
    def get_cmake_components(self, package):
        opened_file = self.read_package_file(package, "cmake-components")
        if opened_file:
            with opened_file as f:
                return [l.rstrip() for l in f if not re.search("^($|#)", l)]
        else:
            return []

    def install_make_arg_parser(self, parser):
        self.make_arg_parser(parser)
        parser.add_argument(
            "--component", action="append",
            help="Component to install for a package")
        parser.add_argument(
            "--sourcedir", action="store",
            help="Source directory for installation (not used except to notify"
                 " dh_missing)",
            default="debian/tmp")

    @common.DHEntryPoint("dh_cmake_install")
    def install(self, args=None):
        self.parse_args(args, make_arg_parser=self.install_make_arg_parser)
        if self.options.component:
            packages = self.get_packages()
            if len(packages) != 1:
                raise common.PackageError("Can only specify one package when "
                                          "specifying components")
            p = packages[0]
            for c in self.options.component:
                self.do_cmake_install(self.get_build_directory(), p,
                                      component=c)
        else:
            for p in self.get_packages():
                for c in self.get_cmake_components(p):
                    self.do_cmake_install(self.get_build_directory(), p,
                                          component=c)


def install():
    dhcmake = DHCMake()
    dhcmake.install()



# This file is part of dh-cmake, and is distributed under the OSI-approved
# BSD 3-Clause license. See top-level LICENSE file or
# https://gitlab.kitware.com/debian/dh-cmake/blob/master/LICENSE for details.

import http.server
import re
import subprocess
import threading
import urllib.parse
import xml.etree.ElementTree
import os

from dhcmake import ctest
from . import DebianSourcePackageTestCaseBase, KWTestCaseBase


class PushEnvironmentVariable:
    def __init__(self, name, value):
        self.name = name
        try:
            self.old_value = os.environ[name]
        except KeyError:
            self.old_value = None

        os.environ[name] = value

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self.old_value is None:
            del os.environ[self.name]
        else:
            os.environ[self.name] = self.old_value


class PushEnvironmentVariableTestCase(KWTestCaseBase):
    varname = "DH_CMAKE_TEST_VARIABLE_DO_NOT_SET"

    def test_create_new(self):
        try:
            del os.environ[self.varname]
        except KeyError:
            pass

        with PushEnvironmentVariable(self.varname, "value"):
            self.assertEqual("value", os.environ[self.varname])

        self.assertNotIn(self.varname, os.environ)

    def test_change(self):
        os.environ[self.varname] = "old"

        with PushEnvironmentVariable(self.varname, "new"):
            self.assertEqual("new", os.environ[self.varname])

        self.assertEqual("old", os.environ[self.varname])


class MockCDashServerHandler(http.server.BaseHTTPRequestHandler):
    def do_PUT(self):
        match = re.search(r"^/submit\.php\?(.*)$", self.path)
        if not match:
            self.send_error(404)
            return
        query_string_params = urllib.parse.parse_qs(match.group(1))

        if query_string_params["project"] != ["dh-cmake-test"]:
            self.send_error(404)
            return

        self.send_response(100)
        self.end_headers()
        self.flush_headers()

        input_file = self.rfile.read(int(self.headers.get("content-length")))
        self.server.submitted_files.add(input_file)

        self.send_response(200)
        self.end_headers()
        self.flush_headers()

    def log_message(self, format, *args):
        pass


class MockCDashServer(http.server.HTTPServer):
    def __init__(self, server_address):
        super().__init__(server_address, MockCDashServerHandler)

        self.submitted_files = set()


class DHCTestTestCase(DebianSourcePackageTestCaseBase):
    DHClass = ctest.DHCTest

    def setUp(self):
        super().setUp()

        try:
            del os.environ["DEB_CTEST_OPTIONS"]
        except KeyError:
            pass  # No variable, no problem

        self.cdash_server = MockCDashServer(("localhost", 47806))
        self.cdash_server_thread = \
            threading.Thread(target=self.cdash_server.serve_forever)
        self.cdash_server_thread.daemon = True
        self.cdash_server_thread.start()

    def tearDown(self):
        self.cdash_server.shutdown()
        self.cdash_server_thread.join()
        self.cdash_server.server_close()

        super().tearDown()

    def assertFilesSubmittedEqual(self, steps):
        contents_set = set()
        for step in steps:
            date = self.get_testing_tag_date()
            with open(os.path.join("debian/.ctest/Testing", date,
                                   step + ".xml"), "rb") as f:
                contents = f.read()
            contents_set.add(contents)

        self.assertEqual(contents_set, self.cdash_server.submitted_files)

    def get_testing_tag_date(self):
        with open("debian/.ctest/Testing/TAG", "r") as f:
            return next(f).rstrip()

    def test_get_deb_ctest_option(self):
        with PushEnvironmentVariable("DEB_CTEST_OPTIONS",
                                     "opt1 opt2=val  opt3=\"spaced value\" opt4=another\\ \\ space"):
            self.assertEqual(True, ctest.get_deb_ctest_option("opt1"))
            self.assertEqual("val", ctest.get_deb_ctest_option("opt2"))
            self.assertEqual(
                "spaced value", ctest.get_deb_ctest_option("opt3"))
            self.assertEqual("another  space",
                             ctest.get_deb_ctest_option("opt4"))

        with PushEnvironmentVariable("DEB_CTEST_OPTIONS",
                                     "opt1=\"a"):
            with self.assertRaisesRegex(ValueError, "Unclosed quote"):
                ctest.get_deb_ctest_option("opt1")

        with PushEnvironmentVariable("DEB_CTEST_OPTIONS",
                                     "opt1=a\\"):
            with self.assertRaisesRegex(ValueError, "Unclosed backslash"):
                ctest.get_deb_ctest_option("opt1")

    def test_clean(self):
        os.makedirs("debian/.ctest/Testing")
        with open("debian/.ctest/Testing/TAG", "w") as f:
            f.write("")
        self.dh.clean([])
        self.assertFileNotExists("debian/.ctest")

    def test_clean_dir(self):
        os.makedirs("debian/ctest/Testing")
        with open("debian/ctest/Testing/TAG", "w") as f:
            f.write("")
        self.dh.clean(["--ctest-testing-dir", "debian/ctest"])
        self.assertFileNotExists("debian/.ctest")

    def test_start_none(self):
        self.assertFileNotExists("debian/.ctest/Testing/TAG")
        self.dh.start([])
        self.assertFileNotExists("debian/.ctest/Testing/TAG")

    def test_start_experimental(self):
        self.assertFileNotExists("debian/.ctest/Testing/TAG")

        with PushEnvironmentVariable("DEB_CTEST_OPTIONS",
                                     "model=Experimental track=\"Experimental Track\""):
            self.dh.start([])
            with open("debian/.ctest/Testing/TAG", "r") as f:
                self.assertRegex(next(f), "^[0-9]{8}-[0-9]{4}$")
                self.assertEqual("Experimental Track", next(f).rstrip())
                try:  # Extra line here as of CMake 3.12
                    self.assertEqual("Experimental", next(f).rstrip())
                except StopIteration:
                    pass
                with self.assertRaises(StopIteration):
                    next(f)

    def test_start_experimental_dir(self):
        self.assertFileNotExists("dummydir/Testing/TAG")

        with PushEnvironmentVariable("DEB_CTEST_OPTIONS",
                                     "model=Experimental"):
            self.dh.start(["--ctest-testing-dir=dummydir"])
            with open("dummydir/Testing/TAG", "r") as f:
                self.assertRegex(next(f), "^[0-9]{8}-[0-9]{4}$")
                self.assertEqual("Experimental", next(f).rstrip())
                try:  # Extra line here as of CMake 3.12
                    self.assertEqual("Experimental", next(f).rstrip())
                except StopIteration:
                    pass
                with self.assertRaises(StopIteration):
                    next(f)

    def test_start_nightly(self):
        self.assertFileNotExists("debian/.ctest/Testing/TAG")

        with PushEnvironmentVariable("DEB_CTEST_OPTIONS", "model=Nightly"):
            self.dh.start([])
            with open("debian/.ctest/Testing/TAG", "r") as f:
                self.assertRegex(next(f), "^[0-9]{8}-[0-9]{4}$")
                self.assertEqual("Nightly", next(f).rstrip())
                try:  # Extra line here as of CMake 3.12
                    self.assertEqual("Nightly", next(f).rstrip())
                except StopIteration:
                    pass
                with self.assertRaises(StopIteration):
                    next(f)

    def test_no_update_experimental(self):
        self.assertFileNotExists("debian/.ctest/Testing/TAG")

        with PushEnvironmentVariable("DEB_CTEST_OPTIONS",
                                     "model=Experimental"):
            self.dh.start([])
            self.dh.update([])
            date = self.get_testing_tag_date()

            self.assertFileNotExists(os.path.join("debian/.ctest/Testing", date,
                                                  "Update.xml"))

    def test_update_experimental(self):
        self.assertFileNotExists("debian/.ctest/Testing/TAG")

        with PushEnvironmentVariable("DEB_CTEST_OPTIONS",
                                     "model=Experimental update"):
            self.run_cmd(["git", "init", "."])
            self.run_cmd(["git", "add", "."])
            self.run_cmd(["git", "commit", "-m", "Initial commit"], env={
                "GIT_AUTHOR_NAME": "Kitware Robot",
                "GIT_AUTHOR_EMAIL": "kwrobot@kitware.com",
                "GIT_AUTHOR_DATE": "2020.01.01T00:00:00",
                "GIT_COMMITTER_NAME": "Kitware Robot",
                "GIT_COMMITTER_EMAIL": "kwrobot@kitware.com",
                "GIT_COMMITTER_DATE": "2020.01.01T00:00:00",
            })
            self.dh.start([])
            self.dh.update([])
            date = self.get_testing_tag_date()

            self.assertFileExists(os.path.join("debian/.ctest/Testing", date,
                                               "Update.xml"))

            with open(os.path.join("debian/.ctest/Testing", date,
                                   "Update.xml"),
                      "r") as f:
                tree = xml.etree.ElementTree.fromstring(f.read())

            update_type = self.get_single_element(tree.findall(
                "UpdateType"))
            self.assertEqual("GIT", update_type.text)
            revision = self.get_single_element(tree.findall(
                "Revision"))
            self.assertEqual(
                subprocess.check_output(
                    ["git", "show", "--quiet", "--format=%H"]).rstrip().decode("utf-8"),
                revision.text)

    def test_update_experimental_modified(self):
        self.assertFileNotExists("debian/.ctest/Testing/TAG")

        with PushEnvironmentVariable("DEB_CTEST_OPTIONS",
                                     "model=Experimental update"):
            self.run_cmd(["git", "init", "."])
            self.run_cmd(["git", "add", "."])
            self.run_cmd(["git", "commit", "-m", "Initial commit"], env={
                "GIT_AUTHOR_NAME": "Kitware Robot",
                "GIT_AUTHOR_EMAIL": "kwrobot@kitware.com",
                "GIT_AUTHOR_DATE": "2020.01.01T00:00:00",
                "GIT_COMMITTER_NAME": "Kitware Robot",
                "GIT_COMMITTER_EMAIL": "kwrobot@kitware.com",
                "GIT_COMMITTER_DATE": "2020.01.01T00:00:00",
            })
            os.unlink("CMakeLists.txt")
            self.dh.start([])
            self.dh.update([])
            date = self.get_testing_tag_date()

            self.assertFileExists(os.path.join("debian/.ctest/Testing", date,
                                               "Update.xml"))

            with open(os.path.join("debian/.ctest/Testing", date,
                                   "Update.xml"),
                      "r") as f:
                tree = xml.etree.ElementTree.fromstring(f.read())

            update_type = self.get_single_element(tree.findall(
                "UpdateType"))
            self.assertEqual("GIT", update_type.text)
            revision = self.get_single_element(tree.findall(
                "Revision"))
            self.assertEqual(
                subprocess.check_output(
                    ["git", "show", "--quiet", "--format=%H"]).rstrip().decode("utf-8"),
                revision.text)

    def test_update_experimental_submit(self):
        self.assertFileNotExists("debian/.ctest/Testing/TAG")

        with PushEnvironmentVariable("DEB_CTEST_OPTIONS",
                                     "model=Experimental update submit"):
            self.run_cmd(["git", "init", "."])
            self.run_cmd(["git", "add", "."])
            self.run_cmd(["git", "commit", "-m", "Initial commit"], env={
                "GIT_AUTHOR_NAME": "Kitware Robot",
                "GIT_AUTHOR_EMAIL": "kwrobot@kitware.com",
                "GIT_AUTHOR_DATE": "2020.01.01T00:00:00",
                "GIT_COMMITTER_NAME": "Kitware Robot",
                "GIT_COMMITTER_EMAIL": "kwrobot@kitware.com",
                "GIT_COMMITTER_DATE": "2020.01.01T00:00:00",
            })
            self.dh.start([])
            self.dh.update([])
            date = self.get_testing_tag_date()

            self.assertFileExists(os.path.join("debian/.ctest/Testing", date,
                                               "Update.xml"))

            with open(os.path.join("debian/.ctest/Testing", date,
                                   "Update.xml"),
                      "r") as f:
                tree = xml.etree.ElementTree.fromstring(f.read())

            update_type = self.get_single_element(tree.findall(
                "UpdateType"))
            self.assertEqual("GIT", update_type.text)
            revision = self.get_single_element(tree.findall(
                "Revision"))
            self.assertEqual(
                subprocess.check_output(
                    ["git", "show", "--quiet", "--format=%H"]).rstrip().decode("utf-8"),
                revision.text)

            self.assertFilesSubmittedEqual({"Update"})

    def test_update_experimental_no_submit(self):
        self.assertFileNotExists("debian/.ctest/Testing/TAG")

        with PushEnvironmentVariable("DEB_CTEST_OPTIONS",
                                     "model=Experimental update submit"):
            self.run_cmd(["git", "init", "."])
            self.run_cmd(["git", "add", "."])
            self.run_cmd(["git", "commit", "-m", "Initial commit"], env={
                "GIT_AUTHOR_NAME": "Kitware Robot",
                "GIT_AUTHOR_EMAIL": "kwrobot@kitware.com",
                "GIT_AUTHOR_DATE": "2020.01.01T00:00:00",
                "GIT_COMMITTER_NAME": "Kitware Robot",
                "GIT_COMMITTER_EMAIL": "kwrobot@kitware.com",
                "GIT_COMMITTER_DATE": "2020.01.01T00:00:00",
            })
            self.dh.start([])
            self.dh.update(["--no-submit"])
            date = self.get_testing_tag_date()

            self.assertFileExists(os.path.join("debian/.ctest/Testing", date,
                                               "Update.xml"))

            with open(os.path.join("debian/.ctest/Testing", date,
                                   "Update.xml"),
                      "r") as f:
                tree = xml.etree.ElementTree.fromstring(f.read())

            update_type = self.get_single_element(tree.findall(
                "UpdateType"))
            self.assertEqual("GIT", update_type.text)
            revision = self.get_single_element(tree.findall(
                "Revision"))
            self.assertEqual(
                subprocess.check_output(
                    ["git", "show", "--quiet", "--format=%H"]).rstrip().decode("utf-8"),
                revision.text)

            self.assertFilesSubmittedEqual({})

    def test_update_experimental_revision(self):
        self.assertFileNotExists("debian/.ctest/Testing/TAG")

        with PushEnvironmentVariable("DEB_CTEST_OPTIONS",
                                     "model=Experimental update revision=1.0"):
            self.dh.start([])
            self.dh.update([])
            date = self.get_testing_tag_date()

            self.assertFileExists(os.path.join("debian/.ctest/Testing", date,
                                               "Update.xml"))

            with open(os.path.join("debian/.ctest/Testing", date,
                                   "Update.xml"),
                      "r") as f:
                tree = xml.etree.ElementTree.fromstring(f.read())

            update_type = self.get_single_element(tree.findall(
                "UpdateType"))
            self.assertEqual("GIT", update_type.text)
            revision = self.get_single_element(tree.findall(
                "Revision"))
            self.assertEqual("1.0", revision.text)

    def test_update_experimental_revision_git(self):
        self.assertFileNotExists("debian/.ctest/Testing/TAG")

        with PushEnvironmentVariable("DEB_CTEST_OPTIONS",
                                     "model=Experimental update revision=1.0"):
            self.run_cmd(["git", "init", "."])
            self.run_cmd(["git", "add", "."])
            self.run_cmd(["git", "commit", "-m", "Initial commit"], env={
                "GIT_AUTHOR_NAME": "Kitware Robot",
                "GIT_AUTHOR_EMAIL": "kwrobot@kitware.com",
                "GIT_AUTHOR_DATE": "2020.01.01T00:00:00",
                "GIT_COMMITTER_NAME": "Kitware Robot",
                "GIT_COMMITTER_EMAIL": "kwrobot@kitware.com",
                "GIT_COMMITTER_DATE": "2020.01.01T00:00:00",
            })
            self.dh.start([])
            self.dh.update([])
            date = self.get_testing_tag_date()

            self.assertFileExists(os.path.join("debian/.ctest/Testing", date,
                                               "Update.xml"))

            with open(os.path.join("debian/.ctest/Testing", date,
                                   "Update.xml"),
                      "r") as f:
                tree = xml.etree.ElementTree.fromstring(f.read())

            update_type = self.get_single_element(tree.findall(
                "UpdateType"))
            self.assertEqual("GIT", update_type.text)
            revision = self.get_single_element(tree.findall(
                "Revision"))
            self.assertEqual("1.0", revision.text)

    def test_configure_none(self):
        self.dh.start([])
        self.dh.configure([])

        self.assertFileNotExists(os.path.join("debian/.ctest/Testing/TAG"))
        self.assertFileExists(os.path.join(self.dh.get_build_directory(),
                                           "CMakeCache.txt"))

    def test_configure_none_bad(self):
        self.dh.start([])
        with self.assertRaises(subprocess.CalledProcessError):
            self.dh.configure(
                ["--", "-DDH_CMAKE_ENABLE_BAD_CONFIGURE:BOOL=ON"])

        self.assertFileNotExists(os.path.join("debian/.ctest/Testing/TAG"))
        self.assertFileExists(os.path.join(self.dh.get_build_directory(),
                                           "CMakeCache.txt"))

    def test_configure_experimental(self):
        with PushEnvironmentVariable("DEB_CTEST_OPTIONS",
                                     "model=Experimental"):
            self.dh.start([])
            self.dh.configure([])
            date = self.get_testing_tag_date()

            self.assertFileExists(os.path.join("debian/.ctest/Testing", date,
                                               "Configure.xml"))

            self.assertFileNotExists(os.path.join("debian/.ctest/Testing",
                                                  date, "Build.xml"))

            self.assertFileNotExists(
                os.path.join(self.dh.get_build_directory(),
                             "testflag.txt"))

            self.assertFilesSubmittedEqual({})

            with open(os.path.join("debian/.ctest/Testing", date,
                                   "Configure.xml"),
                      "r") as f:
                tree = xml.etree.ElementTree.fromstring(f.read())

            self.assertEqual("(empty)", tree.attrib["Name"])
            self.assertEqual("(empty)", tree.attrib["BuildName"])

    def test_configure_experimental_site_build_names(self):
        with PushEnvironmentVariable(
                "DEB_CTEST_OPTIONS",
                "model=Experimental site=debtest build=debian-cmake"):
            self.dh.start(["--ctest-build-suffix=-test"])
            self.dh.configure([])
            date = self.get_testing_tag_date()

            with open(os.path.join("debian/.ctest/Testing", date,
                                   "Configure.xml"),
                      "r") as f:
                tree = xml.etree.ElementTree.fromstring(f.read())

            self.assertEqual("debtest", tree.attrib["Name"])
            self.assertEqual("debian-cmake-test", tree.attrib["BuildName"])

    def test_configure_experimental_site_build_names_arg(self):
        with PushEnvironmentVariable("DEB_CTEST_OPTIONS",
                                     "model=Experimental site=debtest"):
            self.dh.start(["--ctest-build=debian-cmake-test"])
            self.dh.configure([])
            date = self.get_testing_tag_date()

            with open(os.path.join("debian/.ctest/Testing", date,
                                   "Configure.xml"),
                      "r") as f:
                tree = xml.etree.ElementTree.fromstring(f.read())

            self.assertEqual("debtest", tree.attrib["Name"])
            self.assertEqual("debian-cmake-test", tree.attrib["BuildName"])

    def test_configure_experimental_submit(self):
        with PushEnvironmentVariable("DEB_CTEST_OPTIONS",
                                     "model=Experimental submit"):
            self.dh.start([])
            self.dh.configure([])
            date = self.get_testing_tag_date()

            self.assertFileExists(os.path.join("debian/.ctest/Testing", date,
                                               "Configure.xml"))

            self.assertFileNotExists(os.path.join("debian/.ctest/Testing",
                                                  date, "Build.xml"))

            self.assertFileNotExists(
                os.path.join(self.dh.get_build_directory(),
                             "testflag.txt"))

            self.assertFilesSubmittedEqual({"Configure"})

    def test_configure_experimental_no_submit(self):
        with PushEnvironmentVariable("DEB_CTEST_OPTIONS",
                                     "model=Experimental submit"):
            self.dh.start([])
            self.dh.configure(["-O--no-submit"])
            date = self.get_testing_tag_date()

            self.assertFileExists(os.path.join("debian/.ctest/Testing", date,
                                               "Configure.xml"))

            self.assertFileNotExists(os.path.join("debian/.ctest/Testing",
                                                  date, "Build.xml"))

            self.assertFileNotExists(
                os.path.join(self.dh.get_build_directory(),
                             "testflag.txt"))

            self.assertFilesSubmittedEqual({})

    def test_configure_experimental_args(self):
        with PushEnvironmentVariable("DEB_CTEST_OPTIONS",
                                     "model=Experimental"):
            self.dh.start([])
            self.dh.configure(["--", "-DDH_CMAKE_TEST_FLAG:BOOL=ON"])
            date = self.get_testing_tag_date()

            self.assertFileExists(os.path.join("debian/.ctest/Testing", date,
                                               "Configure.xml"))

            self.assertFileNotExists(os.path.join("debian/.ctest/Testing",
                                                  date, "Build.xml"))

            self.assertFileExists(
                os.path.join(self.dh.get_build_directory(),
                             "testflag.txt"))

    def test_configure_experimental_bad(self):
        with PushEnvironmentVariable("DEB_CTEST_OPTIONS",
                                     "model=Experimental"):
            self.dh.start([])
            with self.assertRaises(subprocess.CalledProcessError):
                self.dh.configure([
                    "--", "-DDH_CMAKE_ENABLE_BAD_CONFIGURE:BOOL=ON"])
            date = self.get_testing_tag_date()

            self.assertFileExists(os.path.join("debian/.ctest/Testing", date,
                                               "Configure.xml"))

            self.assertFileNotExists(os.path.join("debian/.ctest/Testing",
                                                  date, "Build.xml"))

            self.assertFilesSubmittedEqual({})

    def test_configure_experimental_bad_submit(self):
        with PushEnvironmentVariable("DEB_CTEST_OPTIONS",
                                     "model=Experimental submit"):
            self.dh.start([])
            with self.assertRaises(subprocess.CalledProcessError):
                self.dh.configure([
                    "--", "-DDH_CMAKE_ENABLE_BAD_CONFIGURE:BOOL=ON"])
            date = self.get_testing_tag_date()

            self.assertFileExists(os.path.join("debian/.ctest/Testing", date,
                                               "Configure.xml"))

            self.assertFileNotExists(os.path.join("debian/.ctest/Testing",
                                                  date, "Build.xml"))

            self.assertFilesSubmittedEqual({"Configure"})

    def test_build_none(self):
        self.dh.start([])
        self.dh.configure([])
        self.dh.build([])

        self.assertFileNotExists(os.path.join("debian/.ctest/Testing/TAG"))
        self.assertFileExists(os.path.join(self.dh.get_build_directory(),
                                           "CMakeCache.txt"))
        self.assertFileExists(os.path.join(self.dh.get_build_directory(),
                                           "libdh-cmake-test.so"))

    def test_build_none_bad(self):
        self.dh.start([])
        self.dh.configure(["--", "-DDH_CMAKE_ENABLE_BAD_BUILD:BOOL=ON"])
        with self.assertRaises(subprocess.CalledProcessError):
            self.dh.build([])

        self.assertFileNotExists(os.path.join("debian/.ctest/Testing/TAG"))
        self.assertFileExists(os.path.join(self.dh.get_build_directory(),
                                           "CMakeCache.txt"))

    def test_build_experimental(self):
        with PushEnvironmentVariable("DEB_CTEST_OPTIONS",
                                     "model=Experimental"):
            self.dh.start([])
            self.dh.configure([])
            self.dh.build([])
            date = self.get_testing_tag_date()

            self.assertFileExists(os.path.join("debian/.ctest/Testing", date,
                                               "Build.xml"))
            self.assertFileNotExists(os.path.join("debian/.ctest/Testing",
                                                  date, "Test.xml"))

            self.assertFilesSubmittedEqual({})

    def test_build_experimental_submit(self):
        with PushEnvironmentVariable("DEB_CTEST_OPTIONS",
                                     "model=Experimental submit"):
            self.dh.start([])
            self.dh.configure([])
            self.dh.build([])
            date = self.get_testing_tag_date()

            self.assertFileExists(os.path.join("debian/.ctest/Testing", date,
                                               "Build.xml"))
            self.assertFileNotExists(os.path.join("debian/.ctest/Testing",
                                                  date, "Test.xml"))

            self.assertFilesSubmittedEqual({"Configure", "Build"})

    def test_build_experimental_no_submit(self):
        with PushEnvironmentVariable("DEB_CTEST_OPTIONS",
                                     "model=Experimental submit"):
            self.dh.start([])
            self.dh.configure([])
            self.dh.build(["-O--no-submit"])
            date = self.get_testing_tag_date()

            self.assertFileExists(os.path.join("debian/.ctest/Testing", date,
                                               "Build.xml"))
            self.assertFileNotExists(os.path.join("debian/.ctest/Testing",
                                                  date, "Test.xml"))

            self.assertFilesSubmittedEqual({"Configure"})

    def test_build_experimental_bad(self):
        with PushEnvironmentVariable("DEB_CTEST_OPTIONS",
                                     "model=Experimental"):
            self.dh.start([])
            self.dh.configure([
                "--", "-DDH_CMAKE_ENABLE_BAD_BUILD:BOOL=ON"])
            with self.assertRaises(subprocess.CalledProcessError):
                self.dh.build([])
            date = self.get_testing_tag_date()

            self.assertFileExists(os.path.join("debian/.ctest/Testing", date,
                                               "Build.xml"))
            self.assertFileNotExists(os.path.join("debian/.ctest/Testing",
                                                  date, "Test.xml"))

            self.assertFilesSubmittedEqual({})

    def test_build_experimental_bad_submit(self):
        with PushEnvironmentVariable("DEB_CTEST_OPTIONS",
                                     "model=Experimental submit"):
            self.dh.start([])
            self.dh.configure([
                "--", "-DDH_CMAKE_ENABLE_BAD_BUILD:BOOL=ON"])
            with self.assertRaises(subprocess.CalledProcessError):
                self.dh.build([])
            date = self.get_testing_tag_date()

            self.assertFileExists(os.path.join("debian/.ctest/Testing", date,
                                               "Build.xml"))
            self.assertFileNotExists(os.path.join("debian/.ctest/Testing",
                                                  date, "Test.xml"))

            self.assertFilesSubmittedEqual({"Configure", "Build"})

    def test_test_none(self):
        self.dh.start([])
        self.dh.configure([])
        self.dh.build([])
        self.dh.test([])

        self.assertFileNotExists(os.path.join("debian/.ctest/Testing/TAG"))

    def test_test_none_bad(self):
        self.dh.start([])
        self.dh.configure(["--", "-DDH_CMAKE_ENABLE_BAD_TEST:BOOL=ON"])
        self.dh.build([])
        with self.assertRaises(subprocess.CalledProcessError):
            self.dh.test([])

        self.assertFileNotExists(os.path.join("debian/.ctest/Testing/TAG"))

    def test_test_none_bad_exclude(self):
        self.dh.start([])
        self.dh.configure(["--", "-DDH_CMAKE_ENABLE_BAD_TEST:BOOL=ON"])
        self.dh.build([])
        self.dh.test(["--", "-E", "TestFalse"])

        self.assertFileNotExists(os.path.join("debian/.ctest/Testing/TAG"))

    def test_test_experimental(self):
        with PushEnvironmentVariable("DEB_CTEST_OPTIONS",
                                     "model=Experimental"):
            self.dh.start([])
            self.dh.configure([])
            self.dh.build([])
            self.dh.test([])
            date = self.get_testing_tag_date()

            with open(os.path.join("debian/.ctest/Testing", date, "Test.xml"),
                      "r") as f:
                tree = xml.etree.ElementTree.fromstring(f.read())

            tests = tree.findall("Testing/Test")
            self.assertEqual(1, len(tests))

            test_true = self.get_single_element(tree.findall(
                "Testing/Test[Name='TestTrue']"))
            self.assertEqual("passed", test_true.get("Status"))

            self.assertFilesSubmittedEqual({})

    def test_test_experimental_submit(self):
        with PushEnvironmentVariable("DEB_CTEST_OPTIONS",
                                     "model=Experimental submit"):
            self.dh.start([])
            self.dh.configure([])
            self.dh.build([])
            self.dh.test([])
            date = self.get_testing_tag_date()

            with open(os.path.join("debian/.ctest/Testing", date, "Test.xml"),
                      "r") as f:
                tree = xml.etree.ElementTree.fromstring(f.read())

            tests = tree.findall("Testing/Test")
            self.assertEqual(1, len(tests))

            test_true = self.get_single_element(tree.findall(
                "Testing/Test[Name='TestTrue']"))
            self.assertEqual("passed", test_true.get("Status"))

            self.assertFilesSubmittedEqual({"Configure", "Build", "Test"})

    def test_test_experimental_no_submit(self):
        with PushEnvironmentVariable("DEB_CTEST_OPTIONS",
                                     "model=Experimental submit"):
            self.dh.start([])
            self.dh.configure([])
            self.dh.build([])
            self.dh.test(["-O--no-submit"])
            date = self.get_testing_tag_date()

            with open(os.path.join("debian/.ctest/Testing", date, "Test.xml"),
                      "r") as f:
                tree = xml.etree.ElementTree.fromstring(f.read())

            tests = tree.findall("Testing/Test")
            self.assertEqual(1, len(tests))

            test_true = self.get_single_element(tree.findall(
                "Testing/Test[Name='TestTrue']"))
            self.assertEqual("passed", test_true.get("Status"))

            self.assertFilesSubmittedEqual({"Configure", "Build"})

    def test_test_experimental_bad(self):
        with PushEnvironmentVariable("DEB_CTEST_OPTIONS",
                                     "model=Experimental"):
            self.dh.start([])
            self.dh.configure([
                "--", "-DDH_CMAKE_ENABLE_BAD_TEST:BOOL=ON"])
            self.dh.build([])
            self.dh.test([])
            date = self.get_testing_tag_date()

            with open(os.path.join("debian/.ctest/Testing", date, "Test.xml"),
                      "r") as f:
                tree = xml.etree.ElementTree.fromstring(f.read())

            tests = tree.findall("Testing/Test")
            self.assertEqual(2, len(tests))

            test_true = self.get_single_element(tree.findall(
                "Testing/Test[Name='TestTrue']"))
            self.assertEqual("passed", test_true.get("Status"))

            test_false = self.get_single_element(tree.findall(
                "Testing/Test[Name='TestFalse']"))
            self.assertEqual("failed", test_false.get("Status"))

            self.assertFilesSubmittedEqual({})

    def test_test_experimental_bad_exclude(self):
        with PushEnvironmentVariable("DEB_CTEST_OPTIONS",
                                     "model=Experimental"):
            self.dh.start([])
            self.dh.configure([
                "--", "-DDH_CMAKE_ENABLE_BAD_TEST:BOOL=ON"])
            self.dh.build([])
            self.dh.test(["--", "-E", "TestFalse"])
            date = self.get_testing_tag_date()

            with open(os.path.join("debian/.ctest/Testing", date, "Test.xml"),
                      "r") as f:
                tree = xml.etree.ElementTree.fromstring(f.read())

            tests = tree.findall("Testing/Test")
            self.assertEqual(1, len(tests))

            test_true = self.get_single_element(tree.findall(
                "Testing/Test[Name='TestTrue']"))
            self.assertEqual("passed", test_true.get("Status"))

            self.assertFilesSubmittedEqual({})

    def test_test_experimental_bad_submit(self):
        with PushEnvironmentVariable("DEB_CTEST_OPTIONS",
                                     "model=Experimental submit"):
            self.dh.start([])
            self.dh.configure([
                "--", "-DDH_CMAKE_ENABLE_BAD_TEST:BOOL=ON"])
            self.dh.build([])
            self.dh.test([])
            date = self.get_testing_tag_date()

            with open(os.path.join("debian/.ctest/Testing", date, "Test.xml"),
                      "r") as f:
                tree = xml.etree.ElementTree.fromstring(f.read())

            tests = tree.findall("Testing/Test")
            self.assertEqual(2, len(tests))

            test_true = self.get_single_element(tree.findall(
                "Testing/Test[Name='TestTrue']"))
            self.assertEqual("passed", test_true.get("Status"))

            test_false = self.get_single_element(tree.findall(
                "Testing/Test[Name='TestFalse']"))
            self.assertEqual("failed", test_false.get("Status"))

            self.assertFilesSubmittedEqual({"Configure", "Build", "Test"})

    def test_test_experimental_bad_catchfailed_submit(self):
        with PushEnvironmentVariable("DEB_CTEST_OPTIONS",
                                     "model=Experimental catchfailed submit"):
            self.dh.start([])
            self.dh.configure([
                "--", "-DDH_CMAKE_ENABLE_BAD_TEST:BOOL=ON"])
            self.dh.build([])
            with self.assertRaises(subprocess.CalledProcessError):
                self.dh.test([])
            date = self.get_testing_tag_date()

            with open(os.path.join("debian/.ctest/Testing", date, "Test.xml"),
                      "r") as f:
                tree = xml.etree.ElementTree.fromstring(f.read())

            tests = tree.findall("Testing/Test")
            self.assertEqual(2, len(tests))

            test_true = self.get_single_element(tree.findall(
                "Testing/Test[Name='TestTrue']"))
            self.assertEqual("passed", test_true.get("Status"))

            test_false = self.get_single_element(tree.findall(
                "Testing/Test[Name='TestFalse']"))
            self.assertEqual("failed", test_false.get("Status"))

            self.assertFilesSubmittedEqual({"Configure", "Build", "Test"})

    def test_submit_none(self):
        self.dh.start([])
        self.dh.configure(["-O--no-submit"])
        self.dh.build(["-O--no-submit"])
        self.dh.test(["-O--no-submit"])
        self.dh.submit([])

        self.assertFilesSubmittedEqual(set())

    def test_submit_experimental_nosubmit(self):
        with PushEnvironmentVariable("DEB_CTEST_OPTIONS",
                                     "model=Experimental"):
            self.dh.start([])
            self.dh.configure(["-O--no-submit"])
            self.dh.build(["-O--no-submit"])
            self.dh.test(["-O--no-submit"])
            self.dh.submit([])

            self.assertFilesSubmittedEqual(set())

    def test_submit_experimental_submit(self):
        with PushEnvironmentVariable("DEB_CTEST_OPTIONS",
                                     "model=Experimental submit"):
            self.dh.start([])
            self.dh.configure(["-O--no-submit"])
            self.dh.build(["-O--no-submit"])
            self.dh.test(["-O--no-submit"])
            self.dh.submit([])

            self.assertFilesSubmittedEqual(
                {"Configure", "Build", "Test", "Done"})

    def test_submit_experimental_submit_parts(self):
        with PushEnvironmentVariable("DEB_CTEST_OPTIONS",
                                     "model=Experimental submit"):
            self.dh.start([])
            self.dh.configure(["-O--no-submit"])
            self.dh.build(["-O--no-submit"])
            self.dh.test(["-O--no-submit"])
            self.dh.submit(["--parts", "Configure", "Build"])

            self.assertFilesSubmittedEqual({"Configure", "Build"})

    def test_run_debian_rules_none(self):
        self.run_debian_rules("build", "ctest")

        self.assertFileNotExists("debian/.ctest/Testing/TAG")
        self.assertFileExists("debian/build/CMakeCache.txt")
        self.assertFilesSubmittedEqual(set())

        self.run_debian_rules("clean", "ctest")
        self.assertFileNotExists("debian/.ctest")

    def test_run_debian_rules_experimental(self):
        with PushEnvironmentVariable("DEB_CTEST_OPTIONS",
                                     "model=Experimental"):
            self.run_debian_rules("build", "ctest")

            self.assertFileExists("debian/.ctest/Testing/TAG")
            self.assertFileExists("debian/build/CMakeCache.txt")
            self.assertFilesSubmittedEqual(set())

            self.run_debian_rules("clean", "ctest")
            self.assertFileNotExists("debian/.ctest")

    def test_run_debian_rules_experimental_submit(self):
        with PushEnvironmentVariable("DEB_CTEST_OPTIONS",
                                     "model=Experimental submit"):
            self.run_debian_rules("build", "ctest")

            self.assertFileExists("debian/.ctest/Testing/TAG")
            self.assertFileExists("debian/build/CMakeCache.txt")
            self.assertFilesSubmittedEqual(
                {"Configure", "Build", "Test", "Done"})

            self.run_debian_rules("clean", "ctest")
            self.assertFileNotExists("debian/.ctest")



# This file is part of dh-cmake, and is distributed under the OSI-approved
# BSD 3-Clause license. See top-level LICENSE file or
# https://gitlab.kitware.com/debian/dh-cmake/blob/master/LICENSE for details.

import autopep8
import pyflakes.api
import os.path
from unittest import TestCase


class SourceCheckTestCase(TestCase):
    def foreach_py(self, func):
        test_dir = os.path.dirname(__file__)
        root_dir = os.path.dirname(os.path.dirname(test_dir))

        for dirpath, _, filenames in os.walk(root_dir):
            for filename in filter(lambda f: f.endswith(".py"), filenames):
                filename_abs = os.path.join(dirpath, filename)
                filename_rel = os.path.relpath(filename_abs, root_dir)
                with open(filename_abs) as f:
                    contents = f.read()
                func(filename_abs, filename_rel, contents)

    def test_autopep8(self):
        self.foreach_py(lambda a, r, c: self.assertEqual(autopep8.fix_code(c), c,
                                                         msg="File %s is incorrectly formatted" % r))

    def test_pyflakes(self):
        self.foreach_py(lambda a, r, c: self.assertEqual(
            0, pyflakes.api.check(c, r), msg="File %s failed pyflakes check" % r))



# This file is part of dh-cmake, and is distributed under the OSI-approved
# BSD 3-Clause license. See top-level LICENSE file or
# https://gitlab.kitware.com/debian/dh-cmake/blob/master/LICENSE for details.

import os.path

from dhcmake import common, cmake
from . import KWTestCaseBase, DebianSourcePackageTestCaseBase


class DHCMakeTestCase(DebianSourcePackageTestCaseBase):
    DHClass = cmake.DHCMake

    libraries_files = set(KWTestCaseBase.replace_arch_in_paths({
        "usr",
        "usr/lib",
        "usr/lib/{arch}",
        "usr/lib/{arch}/libdh-cmake-test.so.1",
        "usr/lib/{arch}/libdh-cmake-test.so.1.0",
        "usr/lib/{arch}/libdh-cmake-test-lib1.so.1",
        "usr/lib/{arch}/libdh-cmake-test-lib1.so.1.0",
        "usr/lib/{arch}/libdh-cmake-test-lib2.so.1",
        "usr/lib/{arch}/libdh-cmake-test-lib2.so.1.0",
    }))

    shlibs_files = {
        "DEBIAN",
        "DEBIAN/shlibs",
    }

    headers_files = set(KWTestCaseBase.replace_arch_in_paths({
        "usr",
        "usr/include",
        "usr/include/dh-cmake-test.h",
        "usr/include/dh-cmake-test-lib1.h",
        "usr/include/dh-cmake-test-lib2.h",
    }))

    namelinks_files = set(KWTestCaseBase.replace_arch_in_paths({
        "usr",
        "usr/lib",
        "usr/lib/{arch}",
        "usr/lib/{arch}/libdh-cmake-test.so",
        "usr/lib/{arch}/libdh-cmake-test-lib1.so",
        "usr/lib/{arch}/libdh-cmake-test-lib2.so",
    }))

    libdh_cmake_test_files = {
        "usr",
        "usr/share",
        "usr/share/doc",
        "usr/share/doc/libdh-cmake-test",
        "usr/share/doc/libdh-cmake-test/changelog.Debian.gz",
    }

    libdh_cmake_test_dev_files = {
        "usr",
        "usr/share",
        "usr/share/doc",
        "usr/share/doc/libdh-cmake-test-dev",
        "usr/share/doc/libdh-cmake-test-dev/changelog.Debian.gz",
    }

    def setup_do_cmake_install(self):
        self.build_dir = self.make_directory_in_tmp("build")

        self.run_cmd(
            [
                "cmake", "-G", "Unix Makefiles", "-DCMAKE_INSTALL_PREFIX=/usr",
                self.src_dir,
            ], cwd=self.build_dir)

        self.run_cmd(["make"], cwd=self.build_dir)

    def test_cmake_install_all(self):
        self.setup_do_cmake_install()
        self.dh.tool_name = "dh_test_cmake_install_all"
        self.dh.parse_args([])
        self.dh.options.sourcedir = "debian/tmp"

        self.dh.do_cmake_install(self.build_dir,
                                 "libdh-cmake-test")

        self.assertFileTreeEqual(self.libraries_files | self.headers_files
                                 | self.namelinks_files, "debian/libdh-cmake-test")

    def test_cmake_install_subdirectory(self):
        self.setup_do_cmake_install()
        self.dh.tool_name = "dh_test_cmake_install_subdirectory"
        self.dh.parse_args([])
        self.dh.options.sourcedir = "debian/tmp"

        self.dh.do_cmake_install(
            self.build_dir,
            "libdh-cmake-test", subdir="lib1")

        self.assertFileTreeEqual(set(self.replace_arch_in_paths({
            "usr",
            "usr/lib",
            "usr/lib/{arch}",
            "usr/lib/{arch}/libdh-cmake-test-lib1.so",
            "usr/lib/{arch}/libdh-cmake-test-lib1.so.1",
            "usr/lib/{arch}/libdh-cmake-test-lib1.so.1.0",
            "usr/include",
            "usr/include/dh-cmake-test-lib1.h",
        })), "debian/libdh-cmake-test")

    def test_cmake_install_one_component(self):
        self.setup_do_cmake_install()
        self.dh.tool_name = "dh_test_cmake_install_one_component"
        self.dh.parse_args([])
        self.dh.options.sourcedir = "debian/tmp"

        self.dh.do_cmake_install(self.build_dir,
                                 "libdh-cmake-test-dev",
                                 component="Headers")

        self.assertFileTreeEqual(
            self.headers_files, "debian/libdh-cmake-test-dev")
        expected_contents = "\n".join(self.replace_arch_in_paths([
            "debian/tmp/usr/include/dh-cmake-test.h",
            "debian/tmp/usr/include/dh-cmake-test-lib1.h",
            "debian/tmp/usr/include/dh-cmake-test-lib2.h",
        ])) + "\n"
        self.assertFileContentsEqual(expected_contents,
                                     "debian/.debhelper/generated/libdh-cmake-test-dev/"
                                     "installed-by-dh_test_cmake_install_one_component")

    def test_get_cmake_components(self):
        self.dh.parse_args([])

        self.assertEqual([
            "Libraries",
        ], self.dh.get_cmake_components("libdh-cmake-test"))

    def test_get_cmake_components_executable(self):
        self.dh.parse_args([])

        self.assertEqual([
            "Headers",
            "Namelinks",
        ], self.dh.get_cmake_components("libdh-cmake-test-dev"))

    def test_get_cmake_components_noexist(self):
        self.dh.parse_args([])

        self.assertEqual([], self.dh.get_cmake_components(
            "libdh-cmake-test-doc"))

    def do_dh_cmake_install(self, args):
        self.dh.parse_args(
            args, make_arg_parser=self.dh.install_make_arg_parser)

        os.mkdir(self.dh.get_build_directory())

        self.run_cmd(
            [
                "cmake", "-G", "Unix Makefiles", "-DCMAKE_INSTALL_PREFIX=/usr",
                self.src_dir,
            ], cwd=self.dh.get_build_directory())

        self.run_cmd(["make"], cwd=self.dh.get_build_directory())

        self.dh.install(args)

    def test_dh_cmake_install_default(self):
        self.do_dh_cmake_install([])

        self.assertFileTreeEqual(self.libraries_files,
                                 "debian/libdh-cmake-test")

        self.assertFileTreeEqual(self.headers_files | self.namelinks_files,
                                 "debian/libdh-cmake-test-dev")

    def test_dh_cmake_install_package_component(self):
        self.do_dh_cmake_install(["--package", "libdh-cmake-test",
                                  "--component", "Namelinks", "--component",
                                  "Libraries"])

        self.assertFileTreeEqual(self.libraries_files | self.namelinks_files,
                                 "debian/libdh-cmake-test")

    def test_dh_cmake_install_package_component_error(self):
        with self.assertRaisesRegex(
                common.PackageError,
                "Can only specify one package when specifying components"):
            self.do_dh_cmake_install(["--package", "libdh-cmake-test",
                                      "--package", "libdh-cmake-test-dev",
                                      "--component", "Namelinks",
                                      "--component", "Libraries"])

    def test_dh_cmake_install_tmpdir(self):
        self.do_dh_cmake_install(["--tmpdir=debian/tmp"])

        self.assertFileTreeEqual(self.libraries_files | self.headers_files
                                 | self.namelinks_files,
                                 "debian/tmp")

    def test_run_debian_rules(self):
        self.run_debian_rules("build", "cmake")
        self.run_debian_rules("install", "cmake")

        self.assertFileTreeEqual(self.libraries_files | self.shlibs_files
                                 | self.libdh_cmake_test_files,
                                 "debian/libdh-cmake-test")

        self.assertFileTreeEqual(self.headers_files | self.namelinks_files
                                 | self.libdh_cmake_test_dev_files,
                                 "debian/libdh-cmake-test-dev")



# This file is part of dh-cmake, and is distributed under the OSI-approved
# BSD 3-Clause license. See top-level LICENSE file or
# https://gitlab.kitware.com/debian/dh-cmake/blob/master/LICENSE for details.

import os.path
import pkg_resources
import re

from dhcmake import common


def format_arg_for_ctest(arg):
    arg = arg.replace("\\", "\\\\")
    arg = arg.replace('"', '\\"')
    arg = arg.replace("'", "\\'")
    arg = arg.replace(" ", "\\ ")
    arg = arg.replace("\t", "\\\t")
    return arg


def format_args_for_ctest(args):
    return " ".join(format_arg_for_ctest(a) for a in args)


def _get_deb_ctest_options():
    try:
        deb_ctest_options = os.environ["DEB_CTEST_OPTIONS"]
    except KeyError:
        return []

    result = []
    current = ""
    escape = False
    quote = False
    for c in deb_ctest_options:
        if escape:
            current += c
            escape = False
        elif c == "\\":
            escape = True
        elif c == '"':
            quote = not quote
        elif re.match("\\s", c) and not quote:
            if current:
                result.append(current)
                current = ""
        else:
            current += c

    if quote:
        raise ValueError("Unclosed quote")
    if escape:
        raise ValueError("Unclosed backslash")

    if current:
        result.append(current)
    return result


def get_deb_ctest_option(name):
    for item in _get_deb_ctest_options():
        eq_split = item.split("=", 1)
        if eq_split[0] == name:
            if len(eq_split) > 1:
                return eq_split[1]
            else:
                return True

    return None


class DHCTest(common.DHCommon):
    def make_arg_parser(self, parser):
        super().make_arg_parser(parser)

        parser.add_argument(
            "--no-submit", action="store_true",
            help="Don't submit after each part")
        parser.add_argument(
            "--ctest-testing-dir", action="store", default="debian/.ctest",
            help="Directory in which to store CTest Testing/ directory")
        parser.add_argument(
            "--ctest-build", action="store",
            help="Name of the CTest build")
        parser.add_argument(
            "--ctest-build-suffix", action="store", default="",
            help="Suffix to add to CTest build")
        parser.add_argument(
            "extra_args", nargs="*")

    def get_dh_ctest_driver(self):
        return pkg_resources.resource_filename(__name__,
                                               "dh_ctest_driver.cmake")

    def do_ctest_step(self, step, cmd=None):
        dashboard_model = get_deb_ctest_option("model")
        if dashboard_model is None:
            if cmd is not None:
                self.do_cmd([cmd, *self.parsed_args])
            return False
        else:
            args = [
                "ctest", "-VV", "-S", self.get_dh_ctest_driver(),
                "-DDH_CTEST_SRCDIR:PATH=" + os.getcwd(),
                "-DDH_CTEST_CTESTDIR:PATH=" + os.path.join(
                    os.getcwd(), self.options.ctest_testing_dir),
                "-DDH_CTEST_BUILDDIR:PATH=" + self.get_build_directory(),
                "-DDH_CTEST_DASHBOARD_MODEL:STRING=" + dashboard_model,
                "-DDH_CTEST_STEP:STRING=" + step,
            ]
            if cmd:
                args.append("-DDH_CTEST_RUN_CMD:STRING="
                            + format_args_for_ctest([cmd, *self.parsed_args]))
            if get_deb_ctest_option("submit") and not self.options.no_submit:
                args.append("-DDH_CTEST_STEP_SUBMIT:BOOL=ON")

            site = get_deb_ctest_option("site")
            if isinstance(site, str):
                args.append("-DDH_CTEST_SITE:STRING=" + site)

            track = get_deb_ctest_option("track")
            if isinstance(track, str):
                args.append("-DDH_CTEST_TRACK:STRING=" + track)

            revision = get_deb_ctest_option("revision")
            if isinstance(revision, str):
                args.append("-DDH_CTEST_VERSION_OVERRIDE:STRING=" + revision)

            build = get_deb_ctest_option("build")
            if self.options.ctest_build:
                build = self.options.ctest_build
            if self.options.ctest_build_suffix:
                build += self.options.ctest_build_suffix
            if isinstance(build, str):
                args.append("-DDH_CTEST_BUILD:STRING=" + build)

            catchfailed = "0"
            if get_deb_ctest_option("catchfailed"):
                catchfailed = "1"
            args.append("-DDH_CTEST_CATCHFAILED:BOOL=" + catchfailed)

            if step == "submit" and self.options.parts:
                args.append("-DDH_CTEST_SUBMIT_PARTS:STRING=" +
                            ";".join(self.options.parts))

            args.extend(self.options.extra_args)
            self.do_cmd(args)
            return True

    @common.DHEntryPoint("dh_ctest_clean")
    def clean(self, args=None):
        self.parse_args(args)
        self.do_cmd(["rm", "-rf", self.options.ctest_testing_dir + "/"])

    @common.DHEntryPoint("dh_ctest_start")
    def start(self, args=None):
        self.parse_args(args)
        self.do_ctest_step("start")

    @common.DHEntryPoint("dh_ctest_update")
    def update(self, args=None):
        self.parse_args(args)
        if get_deb_ctest_option("update"):
            self.do_ctest_step("update")

    @common.DHEntryPoint("dh_ctest_configure")
    def configure(self, args=None):
        self.parse_args(args)
        self.do_ctest_step("configure", "dh_auto_configure")

    @common.DHEntryPoint("dh_ctest_build")
    def build(self, args=None):
        self.parse_args(args)
        self.do_ctest_step("build", "dh_auto_build")

    @common.DHEntryPoint("dh_ctest_test")
    def test(self, args=None):
        self.parse_args(args)
        if not self.do_ctest_step("test"):
            self.do_cmd(["ctest", "-VV", *self.options.extra_args],
                        cwd=self.get_build_directory())

    def submit_make_arg_parser(self, parser):
        super().make_arg_parser(parser)

        parser.add_argument(
            "--ctest-testing-dir", action="store", default="debian/.ctest",
            help="Directory in which to store CTest Testing/ directory")
        parser.add_argument(
            "--ctest-build", action="store",
            help="Name of the CTest build")
        parser.add_argument(
            "--ctest-build-suffix", action="store", default="",
            help="Suffix to add to CTest build")
        parser.add_argument("--parts", action="store", nargs="*",
                            help="Parts to submit to CDash")

    @common.DHEntryPoint("dh_ctest_submit")
    def submit(self, args=None):
        self.parse_args(args, make_arg_parser=self.submit_make_arg_parser)
        self.options.no_submit = False
        self.options.extra_args = []
        if get_deb_ctest_option("submit"):
            self.do_ctest_step("submit")


def clean():
    dhctest = DHCTest()
    dhctest.clean()


def start():
    dhctest = DHCTest()
    dhctest.start()


def update():
    dhctest = DHCTest()
    dhctest.update()


def configure():
    dhctest = DHCTest()
    dhctest.configure()


def build():
    dhctest = DHCTest()
    dhctest.build()


def test():
    dhctest = DHCTest()
    dhctest.test()


def submit():
    dhctest = DHCTest()
    dhctest.submit()



# This file is part of dh-cmake, and is distributed under the OSI-approved
# BSD 3-Clause license. See top-level LICENSE file or
# https://gitlab.kitware.com/debian/dh-cmake/blob/master/LICENSE for details.

#from unittest import skip

import os
from dhcmake import common, arch
from . import DebianSourcePackageTestCaseBase, VolatileNamedTemporaryFile


class DHCommonTestCase(DebianSourcePackageTestCaseBase):
    DHClass = common.DHCommon

    def setUp(self):
        super().setUp()
        self.deb_host_arch_old = arch.dpkg_architecture()["DEB_HOST_ARCH"]
        arch.dpkg_architecture()["DEB_HOST_ARCH"] = "armhf"

    def tearDown(self):
        arch.dpkg_architecture()["DEB_HOST_ARCH"] = self.deb_host_arch_old
        super().tearDown()

    def check_packages(self, expected_packages):
        self.assertEqual(expected_packages, set(self.dh.get_packages()))

    def strip_dh_cmake_compat(self):
        contents = ""
        with open("debian/control", "r") as f:
            for l in f:
                if l != "    dh-cmake-compat (= 1),\n":
                    contents += l
        with open("debian/control", "w") as f:
            f.write(contents)

    def test_do_cmd(self):
        self.dh.parse_args([])

        with VolatileNamedTemporaryFile() as f:
            self.dh.do_cmd(["rm", f.name])
            self.assertVolatileFileNotExists(f.name)

    def test_do_cmd_no_act(self):
        self.dh.parse_args(["--no-act"])

        with VolatileNamedTemporaryFile() as f:
            self.dh.do_cmd(["rm", f.name])
            self.assertFileExists(f.name)

    def test_get_packages_default(self):
        self.dh.parse_args([])

        self.check_packages({
            "libdh-cmake-test",
            "libdh-cmake-test-dev",
            "libdh-cmake-test-doc",
            "libdh-cmake-test-extra-32",
            "libdh-cmake-test-extra-both",
        })

    def test_get_packages_whitelist_short(self):
        self.dh.parse_args(["-plibdh-cmake-test-dev",
                            "-plibdh-cmake-test-doc"])

        self.check_packages({
            "libdh-cmake-test-dev",
            "libdh-cmake-test-doc",
        })

    def test_get_packages_whitelist_long(self):
        self.dh.parse_args(["--package", "libdh-cmake-test-dev",
                            "--package", "libdh-cmake-test-doc"])

        self.check_packages({
            "libdh-cmake-test-dev",
            "libdh-cmake-test-doc",
        })

    def test_get_packages_blacklist_short(self):
        self.dh.parse_args(["-Nlibdh-cmake-test-dev",
                            "-Nlibdh-cmake-test-doc"])

        self.check_packages({
            "libdh-cmake-test",
            "libdh-cmake-test-extra-32",
            "libdh-cmake-test-extra-both",
        })

    def test_get_packages_blacklist_long(self):
        self.dh.parse_args(["--no-package", "libdh-cmake-test-dev",
                            "--no-package", "libdh-cmake-test-doc"])

        self.check_packages({
            "libdh-cmake-test",
            "libdh-cmake-test-extra-32",
            "libdh-cmake-test-extra-both",
        })

    def test_get_packages_arch_short(self):
        self.dh.parse_args(["-a"])

        self.check_packages({
            "libdh-cmake-test",
            "libdh-cmake-test-dev",
            "libdh-cmake-test-extra-32",
            "libdh-cmake-test-extra-both",
        })

    def test_get_packages_arch_long(self):
        self.dh.parse_args(["--arch"])

        self.check_packages({
            "libdh-cmake-test",
            "libdh-cmake-test-dev",
            "libdh-cmake-test-extra-32",
            "libdh-cmake-test-extra-both",
        })

    def test_get_packages_arch_deprecated(self):
        self.dh.parse_args(["-s"])

        self.check_packages({
            "libdh-cmake-test",
            "libdh-cmake-test-dev",
            "libdh-cmake-test-extra-32",
            "libdh-cmake-test-extra-both",
        })

    def test_get_packages_indep_short(self):
        self.dh.parse_args(["-i"])

        self.check_packages({
            "libdh-cmake-test-doc",
        })

    def test_get_packages_indep_long(self):
        self.dh.parse_args(["--indep"])

        self.check_packages({
            "libdh-cmake-test-doc",
        })

    def test_get_main_package_default(self):
        self.dh.parse_args([])

        self.assertEqual("libdh-cmake-test",
                         self.dh.get_main_package())

    def test_get_main_package_specified(self):
        self.dh.parse_args(["--mainpackage=libdh-cmake-test-dev"])

        self.assertEqual("libdh-cmake-test-dev",
                         self.dh.get_main_package())

    def test_get_package_file(self):
        self.dh.parse_args([])

        self.assertEqual(
            "debian/cmake-components",
            self.dh.get_package_file(
                "libdh-cmake-test", "cmake-components"
            )
        )

        self.assertEqual(
            "debian/libdh-cmake-test-dev.cmake-components",
            self.dh.get_package_file(
                "libdh-cmake-test-dev", "cmake-components"
            )
        )

        self.assertIsNone(self.dh.get_package_file(
            "libdh-cmake-test-doc", "cmake-components"))

        self.assertEqual(
            "debian/libdh-cmake-test.specific",
            self.dh.get_package_file(
                "libdh-cmake-test", "specific"
            )
        )

        self.assertEqual(
            "debian/libdh-cmake-test.both",
            self.dh.get_package_file(
                "libdh-cmake-test", "both"
            )
        )

    def test_read_package_file(self):
        self.dh.parse_args([])

        with self.dh.read_package_file(
                "libdh-cmake-test", "cmake-components") as f:
            self.assertEqual(
                "# This file tests both comments and blank lines\n\n"
                "Libraries\n", f.read())
        with self.dh.read_package_file(
                "libdh-cmake-test-dev", "cmake-components") as f:
            self.assertEqual("Headers\nNamelinks\n", f.read())
        self.assertIsNone(self.dh.read_package_file(
            "libdh-cmake-test-doc", "cmake-components"))

    def test_build_directory_default(self):
        self.dh.parse_args([])

        self.assertEqual(
            "obj-" + arch.dpkg_architecture()["DEB_HOST_GNU_TYPE"],
            self.dh.get_build_directory())

    def test_build_directory_short(self):
        self.dh.parse_args(["-B", "debian/build"])

        self.assertEqual("debian/build",
                         self.dh.get_build_directory())

    def test_build_directory_long(self):
        self.dh.parse_args(["--builddirectory", "debian/build"])

        self.assertEqual("debian/build",
                         self.dh.get_build_directory())

    def test_tmpdir_default(self):
        self.dh.parse_args([])

        self.assertEqual("debian/libdh-cmake-test",
                         self.dh.get_tmpdir("libdh-cmake-test"))
        self.assertEqual("debian/libdh-cmake-test-dev",
                         self.dh.get_tmpdir("libdh-cmake-test-dev"))

    def test_tmpdir_short(self):
        self.dh.parse_args(["-P", "debian/tmpdir"])

        self.assertEqual("debian/tmpdir",
                         self.dh.get_tmpdir("libdh-cmake-test"))
        self.assertEqual("debian/tmpdir",
                         self.dh.get_tmpdir("libdh-cmake-test-dev"))

    def test_tmpdir_long(self):
        self.dh.parse_args(["--tmpdir=debian/tmpdir"])

        self.assertEqual("debian/tmpdir",
                         self.dh.get_tmpdir("libdh-cmake-test"))
        self.assertEqual("debian/tmpdir",
                         self.dh.get_tmpdir("libdh-cmake-test-dev"))

    def test_o_flag(self):
        self.dh.parse_args(["-O=-v"])

        self.assertTrue(self.dh.options.verbose)

    def test_compat_valid(self):
        self.assertEqual(1, self.dh.compat())

    def test_compat_low(self):
        self.strip_dh_cmake_compat()
        with open("debian/dh-cmake.compat", "w") as f:
            print(common.MIN_COMPAT - 1, file=f)

        with self.assertRaisesRegex(
            common.CompatError,
            r"Compat level %i too old \(must be %i or newer\)"
                % (common.MIN_COMPAT - 1, common.MIN_COMPAT)):
            self.dh.compat()

    def test_compat_high(self):
        self.strip_dh_cmake_compat()
        with open("debian/dh-cmake.compat", "w") as f:
            print(common.MAX_COMPAT + 1, file=f)

        with self.assertRaisesRegex(
            common.CompatError,
            r"Compat level %i too new \(must be %i or older\)"
                % (common.MAX_COMPAT + 1, common.MAX_COMPAT)):
            self.dh.compat()

    def test_compat_missing(self):
        self.strip_dh_cmake_compat()
        os.unlink("debian/dh-cmake.compat")

        with self.assertRaisesRegex(
                common.CompatError,
                "No compat level specified"):
            self.dh.compat()

    def test_compat_conflicting(self):
        with open("debian/dh-cmake.compat", "w") as f:
            print(2, file=f)

        with self.assertRaisesRegex(
                common.CompatError,
                "Conflicting compat levels: 1, 2"):
            self.dh.compat()


class DHCommonCompatTestClass(common.DHCommon):
    @common.DHEntryPoint("dh_common_test_command")
    def test_command(self, args=None):
        self.parse_args(args)
        # Do nothing


class DHCommonCompatTestCase(DebianSourcePackageTestCaseBase):
    DHClass = DHCommonCompatTestClass

    def test_compat_load(self):
        self.assertIs(None, self.dh._compat)
        self.dh.test_command([])
        self.assertEqual(1, self.dh._compat)
        self.assertEqual("dh_common_test_command", self.dh.tool_name)



# This file is part of dh-cmake, and is distributed under the OSI-approved
# BSD 3-Clause license. See top-level LICENSE file or
# https://gitlab.kitware.com/debian/dh-cmake/blob/master/LICENSE for details.

import argparse
import io
import os.path
import subprocess
import sys

from dhcmake import deb822, arch
import debian.deb822


MIN_COMPAT = 1
MAX_COMPAT = 1


class CompatError(Exception):
    pass


class PackageError(Exception):
    pass


def format_arg_for_print(arg):
    arg = arg.replace("\\", "\\\\")
    arg = arg.replace('"', '\\"')
    arg = arg.replace("'", "\\'")
    arg = arg.replace("$", "\\$")
    if " " in arg:
        return '"%s"' % arg
    else:
        return arg


def escape_substvar(value):
    value = value.replace("$", "${}")
    value = value.replace("\n", "${Newline}")

    return value


def DHEntryPoint(tool_name):
    def wrapper(func):
        def wrapped(self, *args, **kargs):
            self.tool_name = tool_name
            self.compat()
            return func(self, *args, **kargs)

        return wrapped
    return wrapper


class DHCommon:
    def __init__(self):
        self.options = argparse.Namespace()
        self.options.type = "both"
        self.stdout = sys.stdout
        self.stdout_b = sys.stdout
        self.stderr = sys.stderr
        self.stderr_b = sys.stderr
        self._compat = None

    def _parse_args(self, parser, args, known):
        if known:
            parser.parse_known_args(args=args, namespace=self.options)
        else:
            parser.parse_args(args=args, namespace=self.options)

        if self.options.options:
            options = self.options.options
            self.options.options = []
            self._parse_args(parser, options, True)

    def _set_compat(self, compat):
        if self._compat is not None and self._compat != compat:
            raise CompatError("Conflicting compat levels: %i, %i" %
                              (self._compat, compat))
        self._compat = compat

    def compat(self):
        if self._compat is None:
            with open("debian/control", "r") as f:
                source, _ = deb822.read_control(f)
            try:
                deps = source["Build-Depends"]
            except KeyError:
                deps = None
            if deps:
                deps_parsed = debian.deb822.PkgRelation.parse_relations(deps)
                for dep in deps_parsed:
                    for subdep in dep:
                        if subdep["name"] == "dh-cmake-compat" and subdep["version"]:
                            sign, version = subdep["version"]
                            if sign == "=":
                                self._set_compat(int(version))

            try:
                with open("debian/dh-cmake.compat", "r") as f:
                    self._set_compat(int(f.read()))
            except FileNotFoundError:
                pass

            if self._compat is None:
                raise CompatError("No compat level specified")

            if self._compat < MIN_COMPAT:
                raise CompatError(
                    "Compat level %i too old (must be %i or newer)" %
                    (self._compat, MIN_COMPAT))
            elif self._compat > MAX_COMPAT:
                raise CompatError(
                    "Compat level %i too new (must be %i or older)" %
                    (self._compat, MAX_COMPAT))

        return self._compat

    def parse_args(self, args=None, make_arg_parser=None):
        parser = argparse.ArgumentParser()
        if make_arg_parser:
            make_arg_parser(parser)
        else:
            self.make_arg_parser(parser)

        if args is None:
            self.parsed_args = sys.argv[1:]
        else:
            self.parsed_args = list(args)

        self._parse_args(parser, args, False)

    def make_arg_parser(self, parser):
        # Required arguments
        parser.add_argument(
            "-v", "--verbose", action="store_true", help="Verbose output")
        parser.add_argument(
            "--no-act", action="store_true",
            help="Dry run (don't actually do anything)")
        parser.add_argument(
            "-a", "-s", "--arch", action="store_const", const="arch",
            dest="type", help="Act on all architecture dependent packages")
        parser.add_argument(
            "-i", "--indep", action="store_const", const="indep",
            dest="type", help="Act on all architecture independent packages")
        parser.add_argument(
            "-p", "--package", action="append",
            help="Act on a specific package or set of packages")
        parser.add_argument(
            "-N", "--no-package", action="append",
            help="Do not act on a specific package or set of packages")
        parser.add_argument(
            "--remaining-packages", action="append",
            help="Do not act on packages which have already been acted on")
        parser.add_argument(
            "--ignore", action="append", help="Ignore a file or set of files")
        parser.add_argument(
            "-P", "--tmpdir", action="store",
            help="Use tmpdir for package build directory")
        parser.add_argument(
            "--mainpackage", action="store",
            help="Change which package is the \"main package\"")
        parser.add_argument(
            "-O", action="append",
            help="Pass additional arguments to called programs",
            dest="options")

        # More arguments
        parser.add_argument(
            "-B", "--builddirectory", action="store",
            help="Build directory for out of source building",
            default="obj-"
            + arch.dpkg_architecture()["DEB_HOST_GNU_TYPE"])

    def print_cmd(self, args, cwd=None):
        if self.options.verbose:
            args = list(args)
            if cwd:
                args = ["cd", cwd, "&&"] + args
            print_args = (format_arg_for_print(a) for a in args)
            print("\t" + " ".join(print_args), file=self.stdout)

    def do_cmd(self, args, env=None, cwd=None):
        self.print_cmd(args, cwd)
        if not self.options.no_act:
            subprocess.run(args, stdout=self.stdout, stderr=self.stderr,
                           env=env, cwd=cwd, check=True)

    def get_all_packages(self):
        with open("debian/control", "r") as f:
            source, packages = deb822.read_control(f)

        result = []

        for p in packages:
            name = p["package"]
            if p.architecture == ["all"]:
                result.append((name, p, "indep"))
            else:
                result.append((name, p, "arch"))

        return result

    def get_compatible_packages(self):
        deb_host_arch = arch.dpkg_architecture()["DEB_HOST_ARCH"]
        result = []

        for name, p, ptype in self.get_all_packages():
            if ptype == "indep" or \
                    arch.debarch_contains(deb_host_arch, p.architecture):
                result.append((name, p, ptype))

        return result

    def get_packages(self):
        result = []

        for name, p, ptype in self.get_compatible_packages():
            if self.options.package:
                if name in self.options.package:
                    result.append(name)
            elif self.options.no_package:
                if name not in self.options.no_package:
                    result.append(name)
            elif self.options.type == "indep":
                if ptype == "indep":
                    result.append(name)
            elif self.options.type == "arch":
                if ptype == "arch":
                    result.append(name)
            elif self.options.type == "both":
                result.append(name)

        return result

    def get_main_package(self):
        if self.options.mainpackage:
            return self.options.mainpackage
        else:
            with open("debian/control", "r") as f:
                source, packages = deb822.read_control(f)

            return packages[0]["package"]

    def get_package_file(self, package, extension):
        paths = [
            os.path.join("debian", package + "." + extension),
        ]

        if package == self.get_main_package():
            paths.append(os.path.join("debian", extension))

        for p in paths:
            if os.path.exists(p):
                return p
        return None

    def read_package_file(self, package, extension):
        path = self.get_package_file(package, extension)
        if path is None:
            return None
        elif os.access(path, os.X_OK):
            return io.StringIO(subprocess.check_output(
                [os.path.abspath(path)]).decode("utf-8"))
        else:
            return open(path, "r")

    def get_build_directory(self):
        return self.options.builddirectory

    def get_tmpdir(self, package):
        if self.options.tmpdir:
            return self.options.tmpdir
        else:
            return os.path.join("debian", package)

    def write_substvar(self, name, value, package=None):
        if package:
            filename = "debian/%s.substvars" % package
        else:
            filename = "debian/substvars"

        line = "%s=%s" % (name, value)
        self.print_cmd(["echo", line, ">>", filename])
        if not self.options.no_act:
            with open(filename, "a") as f:
                print(line, file=f)

    def log_installed_files(self, package, paths):
        if self.options.no_act:
            return
        pkgdir = "debian/.debhelper/generated/%s" % package
        os.makedirs(pkgdir, exist_ok=True)
        with open(os.path.join(pkgdir, "installed-by-%s" % self.tool_name),
                  "a") as f:
            for p in paths:
                f.write("%s\n" % p)

    def do_cmake_install(self, builddir, package, component=None, subdir=None,
                         extra_args=None):
        build_subdir = builddir
        if subdir:
            build_subdir = os.path.join(builddir, subdir)
        args = ["cmake", "--install", build_subdir]
        if component:
            args += ["--component", component]
        if extra_args:
            args += extra_args
        env = os.environ.copy()
        env["DESTDIR"] = os.path.abspath(self.get_tmpdir(package))
        self.do_cmd(args, env=env)

        if component:
            install_manifest = "install_manifest_%s.txt" % component
        else:
            install_manifest = "install_manifest.txt"
        have_manifest = True
        try:
            with open(os.path.join(builddir, install_manifest)) as f:
                files = [os.path.join(self.options.sourcedir,
                                      os.path.relpath(l.rstrip("\n"), "/"))
                         for l in f]
        except FileNotFoundError:
            have_manifest = False
        if have_manifest:
            os.unlink(os.path.join(builddir, install_manifest))
            self.log_installed_files(package, files)
